@@ -1,10 +1,7 @@
 #[macro_use]
 extern crate lazy_static;
 
-#[macro_use]
-extern crate nanoid;
-
-mod data;
+pub mod data;
 
 pub mod builtins;
 pub mod call_stack;
@@ -21,7 +18,9 @@ use std::cell::RefCell;
 use std::fs;
 use std::path::Path;
 
-pub use primes::{Calcit, CalcitItems};
+use im_ternary_tree::TernaryTreeList;
+
+pub use primes::{Calcit, CalcitErr, CalcitItems};
 
 pub fn load_core_snapshot() -> Result<snapshot::Snapshot, String> {
   // load core libs
@@ -31,47 +30,51 @@ pub fn load_core_snapshot() -> Result<snapshot::Snapshot, String> {
   snapshot::load_snapshot_data(core_data, "calcit-internal://calcit-core.cirru")
 }
 
-pub fn run_program(
-  init_fn: &str,
-  params: CalcitItems,
-  program_code: &program::ProgramCodeData,
-) -> Result<Calcit, String> {
+pub fn run_program(init_fn: &str, params: CalcitItems) -> Result<Calcit, CalcitErr> {
   let (init_ns, init_def) = util::string::extract_ns_def(init_fn)?;
 
   let check_warnings: &RefCell<Vec<String>> = &RefCell::new(vec![]);
 
   // preprocess to init
-  match runner::preprocess::preprocess_ns_def(&init_ns, &init_def, &program_code, &init_def, None, check_warnings) {
+  match runner::preprocess::preprocess_ns_def(
+    &init_ns.to_owned().into_boxed_str(),
+    &init_def.to_owned().into_boxed_str(),
+    &init_def.to_owned().into_boxed_str(),
+    None,
+    check_warnings,
+    &TernaryTreeList::Empty,
+  ) {
     Ok(_) => (),
     Err(failure) => {
       println!("\nfailed preprocessing, {}", failure);
-      call_stack::display_stack(&failure)?;
-      return Err(failure);
+      call_stack::display_stack(&failure.msg, &failure.stack)?;
+      return CalcitErr::err_str(failure.msg);
     }
   }
 
   let warnings = check_warnings.to_owned().into_inner();
   if !warnings.is_empty() {
-    for message in &warnings {
-      println!("{}", message);
-    }
-    return Err(format!("Found {} warnings, runner blocked", warnings.len()));
+    return Err(CalcitErr {
+      msg: format!("Found {} warnings, runner blocked", warnings.len()),
+      warnings: warnings.to_owned(),
+      stack: TernaryTreeList::Empty,
+    });
   }
   match program::lookup_evaled_def(&init_ns, &init_def) {
-    None => Err(format!("entry not initialized: {}/{}", init_ns, init_def)),
+    None => CalcitErr::err_str(format!("entry not initialized: {}/{}", init_ns, init_def)),
     Some(entry) => match entry {
       Calcit::Fn(_, f_ns, _, def_scope, args, body) => {
-        let result = runner::run_fn(&params, &def_scope, &args, &body, &f_ns, &program_code);
+        let result = runner::run_fn(&params, &def_scope, &args, &body, &f_ns, &TernaryTreeList::Empty);
         match result {
           Ok(v) => Ok(v),
           Err(failure) => {
             println!("\nfailed, {}", failure);
-            call_stack::display_stack(&failure)?;
+            call_stack::display_stack(&failure.msg, &failure.stack)?;
             Err(failure)
           }
         }
       }
-      _ => Err(format!("expected function entry, got: {}", entry)),
+      _ => CalcitErr::err_str(format!("expected function entry, got: {}", entry)),
     },
   }
 }
